import logging
from fastapi import FastAPI, Request, logger
from fastapi.middleware.httpsredirect import HTTPSRedirectMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.routing import APIRoute
from starlette.middleware.cors import CORSMiddleware
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from blocks_genesis._cache.cache_provider import CacheProvider
from blocks_genesis._cache.redis_client import RedisClient
from blocks_genesis._core.secret_loader import SecretLoader
from blocks_genesis._database.db_context import DbContext
from blocks_genesis._database.mongo_context import MongoDbContextProvider
from blocks_genesis._lmt.log_config import configure_logger
from blocks_genesis._lmt.mongo_log_exporter import MongoHandler
from blocks_genesis._lmt.tracing import configure_tracing
from blocks_genesis._message.azure.azure_message_client import AzureMessageClient
from blocks_genesis._message.message_configuration import MessageConfiguration
from blocks_genesis._middlewares.global_exception_middleware import GlobalExceptionHandlerMiddleware
from blocks_genesis._middlewares.tenant_middleware import TenantValidationMiddleware
from uvicorn.middleware.proxy_headers import ProxyHeadersMiddleware
from blocks_genesis._tenant.tenant_service import initialize_tenant_service
from fastapi.openapi.docs import get_swagger_ui_html
from fastapi.openapi.utils import get_openapi

logger = logging.getLogger(__name__)

async def configure_lifespan(name: str, message_config: MessageConfiguration):
    logger.info("Initializing services...")
    logger.info("Loading secrets before app creation...")
    secret_loader = SecretLoader(name)
    await secret_loader.load_secrets()
    logger.info("Secrets loaded successfully!")
    
    configure_logger()
    logger.info("Logger started")

    # Enable tracing after secrets are loaded
    configure_tracing()
    logger.info("Tracing enabled successfully!")

    CacheProvider.set_client(RedisClient())
    await initialize_tenant_service()
    DbContext.set_provider(MongoDbContextProvider())
    
    AzureMessageClient.initialize(message_config)



def custom_generate_unique_id(route: APIRoute):
    """
    Custom function to generate unique IDs for routes.
    This is useful for debugging and logging purposes.
    """
    return f"{route.name}-{route.path.replace('/', '_')}"

def fast_api_app(lifespan, **kwargs: FastAPI) -> FastAPI:
    app = FastAPI(
        lifespan=lifespan,
        generate_unique_id_function=custom_generate_unique_id,
        **kwargs
    )
    
    return app
   
    
async def close_lifespan():
    logger.info("Shutting down services...")
    
    await AzureMessageClient.get_instance().close()
    # Shutdown logic
    if hasattr(MongoHandler, '_mongo_logger') and MongoHandler._mongo_logger:
        MongoHandler._mongo_logger.stop()
        
def configure_middlewares(app: FastAPI, show_docs: bool = False):
    app.add_middleware(ProxyHeadersMiddleware, trusted_hosts=["*"])    
    app.add_middleware(GZipMiddleware)
    app.add_middleware(TenantValidationMiddleware)
    app.add_middleware(GlobalExceptionHandlerMiddleware)
    FastAPIInstrumentor.instrument_app(app)  ### Instrument FastAPI for OpenTelemetry
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    
    @app.get("/ping", include_in_schema=False)
    async def health():
        return {
            "status": "healthy",
            "message": "pong",
        }
        
    @app.get("/swagger/index.html", include_in_schema=False)
    async def get_documentation(request:Request):
        root_path = request.scope.get("root_path", None)
        openapi_url = f"{root_path}/openapi.json" if root_path else "/openapi.json"
        print(openapi_url)
        if show_docs:
            return get_swagger_ui_html(openapi_url=openapi_url, title="Swagger")
        else:
            return "NOT_ALLOWED"

    @app.get("/openapi.json", include_in_schema=False)
    async def openapi():
        if show_docs:
            return get_openapi(title=app.title, version=app.version, routes=app.routes)
        return {}
    


import pytest
from datetime import datetime
from blocks_genesis._auth.blocks_context import BlocksContext, BlocksContextManager

@pytest.fixture(autouse=True)
def clear_context():
    BlocksContextManager.clear_context()
    yield
    BlocksContextManager.clear_context()

def test_blocks_context_defaults():
    ctx = BlocksContext()
    assert ctx.tenant_id == ""
    assert ctx.roles == []
    assert ctx.is_authenticated is False

def test_blocks_contextmanager_create_from_jwt_claims():
    claims = {
        'tenant_id': 'tid',
        'roles': ['admin'],
        'user_id': 'uid',
        'request_uri': '/uri',
        'org_id': 'oid',
        'exp': int(datetime.now().timestamp()),
        'email': 'e@e.com',
        'permissions': ['perm'],
        'user_name': 'uname',
        'phone': '123',
        'name': 'disp',
        'oauth': 'token',
    }
    ctx = BlocksContextManager.create_from_jwt_claims(claims)
    assert ctx.tenant_id == 'tid'
    assert ctx.roles == ['admin']
    assert ctx.user_id == 'uid'
    assert ctx.organization_id == 'oid'
    assert ctx.email == 'e@e.com'
    assert ctx.permissions == ['perm']
    assert ctx.user_name == 'uname'
    assert ctx.phone_number == '123'
    assert ctx.display_name == 'disp'
    assert ctx.oauth_token == 'token'
    assert ctx.is_authenticated is True

def test_blocks_contextmanager_create():
    ctx = BlocksContextManager.create(tenant_id='tid', roles=['r'], user_id='uid', is_authenticated=True)
    assert ctx.tenant_id == 'tid'
    assert ctx.roles == ['r']
    assert ctx.user_id == 'uid'
    assert ctx.is_authenticated is True

def test_blocks_contextmanager_set_and_get_context():
    ctx = BlocksContextManager.create(tenant_id='tid')
    BlocksContextManager.set_context(ctx)
    assert BlocksContextManager.get_context().tenant_id == 'tid'

def test_blocks_contextmanager_clear_context():
    ctx = BlocksContextManager.create(tenant_id='tid')
    BlocksContextManager.set_context(ctx)
    BlocksContextManager.clear_context()
    assert BlocksContextManager.get_context() is None

def test_blocks_contextmanager_test_mode():
    BlocksContextManager.set_test_mode(True)
    ctx = BlocksContextManager.create(tenant_id='tid')
    assert BlocksContextManager.get_test_mode() is True
    assert BlocksContextManager.get_context(test_value=ctx).tenant_id == 'tid'
    BlocksContextManager.set_test_mode(False)
    assert BlocksContextManager.get_test_mode() is False 


import pytest
from unittest.mock import MagicMock
from blocks_genesis._database.db_context import DbContext

def test_set_and_get_provider():
    provider = MagicMock()
    DbContext.set_provider(provider)
    assert DbContext.get_provider() is provider

def test_clear_provider():
    provider = MagicMock()
    DbContext.set_provider(provider)
    DbContext.clear()
    with pytest.raises(RuntimeError):
        DbContext.get_provider() 


import pytest
from unittest.mock import AsyncMock, patch, MagicMock
from blocks_genesis._core import change_context

@pytest.mark.asyncio
@patch('blocks_genesis._core.change_context.BlocksContextManager')
@patch('blocks_genesis._core.change_context.Activity')
@patch('blocks_genesis._core.change_context.get_tenant_service')
async def test_change_context_noop(mock_get_tenant_service, mock_activity, mock_context_mgr):
    context = MagicMock()
    context.tenant_id = 'tid'
    mock_context_mgr.get_context.return_value = context
    # project_key is None
    await change_context.change_context(None)
    # project_key is empty
    await change_context.change_context("")
    # project_key is same as tenant_id
    await change_context.change_context('tid')
    # Should not change context
    mock_context_mgr.set_context.assert_not_called()

@pytest.mark.asyncio
@patch('blocks_genesis._core.change_context.BlocksContextManager')
@patch('blocks_genesis._core.change_context.Activity')
@patch('blocks_genesis._core.change_context.get_tenant_service')
async def test_change_context_root(mock_get_tenant_service, mock_activity, mock_context_mgr):
    context = MagicMock()
    context.tenant_id = 'tid'
    context.roles = ['r']
    context.user_id = 'u'
    context.is_authenticated = True
    context.request_uri = '/uri'
    context.organization_id = 'org'
    context.expire_on = None
    context.email = 'e@e.com'
    context.permissions = ['perm']
    context.user_name = 'uname'
    context.phone_number = '123'
    context.display_name = 'disp'
    context.oauth_token = 'token'
    mock_context_mgr.get_context.return_value = context
    tenant = MagicMock()
    tenant._id = True
    mock_get_tenant_service.return_value.get_tenant.return_value = tenant
    await change_context.change_context('newtid')
    mock_context_mgr.set_context.assert_called() 


import pytest
from blocks_genesis._cache.cache_provider import CacheProvider

class DummyCache:
    pass

def test_set_and_get_client():
    dummy = DummyCache()
    CacheProvider.set_client(dummy)
    assert CacheProvider.get_client() is dummy

def test_clear_client():
    dummy = DummyCache()
    CacheProvider.set_client(dummy)
    CacheProvider.clear()
    with pytest.raises(RuntimeError):
        CacheProvider.get_client() 


from blocks_genesis._auth.blocks_context import BlocksContextManager
from blocks_genesis._lmt.activity import Activity
from blocks_genesis._tenant.tenant_service import get_tenant_service


async def change_context(project_key: str):
    context = BlocksContextManager.get_context()
    
    Activity.set_current_property("baggage.ActualTenantId", context.tenant_id)
    
    if project_key is None or project_key == "" or project_key == context.tenant_id:
        return
    
    is_root = (await get_tenant_service().get_tenant(context.tenant_id)).is_root_tenant
    
    if is_root:
        BlocksContextManager.set_context(
            BlocksContextManager.create(
                tenant_id=project_key,
                roles=context.roles,
                user_id=context.user_id,
                is_authenticated=context.is_authenticated,
                request_uri=context.request_uri,
                organization_id=context.organization_id,
                expire_on=context.expire_on,
                email=context.email,
                permissions=context.permissions,
                user_name=context.user_name,
                phone_number=context.phone_number,
                display_name=context.display_name,
                oauth_token=context.oauth_token,
                actual_tenant_id=context.tenant_id
            )
        )
        
        Activity.set_current_property("baggage.TenantId", project_key)
        
        


from blocks_genesis._tenant.tenant import CertificateStorageType, JwtTokenParameters, Tenant
from datetime import datetime

def test_certificate_storage_type_enum():
    assert CertificateStorageType.AZURE == 1
    assert CertificateStorageType.FILESYSTEM == 2
    assert CertificateStorageType.MONGODB == 3

def test_jwt_token_parameters_fields():
    params = JwtTokenParameters(
        Issuer='issuer',
        Subject='subject',
        Audiences=['aud'],
        PublicCertificatePath='path',
        PublicCertificatePassword='pubpass',
        PrivateCertificatePassword='privpass',
        CertificateStorageType=CertificateStorageType.AZURE,
        CertificateValidForNumberOfDays=10,
        IssueDate=datetime.now()
    )
    assert params.issuer == 'issuer'
    assert params.subject == 'subject'
    assert params.audiences == ['aud']
    assert params.public_certificate_path == 'path'
    assert params.public_certificate_password == 'pubpass'
    assert params.private_certificate_password == 'privpass'
    assert params.certificate_storage_type == CertificateStorageType.AZURE
    assert params.certificate_valid_for_number_of_days == 10
    assert isinstance(params.issue_date, datetime)

def test_tenant_model_fields():
    params = JwtTokenParameters(
        Issuer='issuer',
        Subject='subject',
        Audiences=['aud'],
        PublicCertificatePath='path',
        PublicCertificatePassword='pubpass',
        PrivateCertificatePassword='privpass',
        CertificateStorageType=CertificateStorageType.AZURE,
        CertificateValidForNumberOfDays=10,
        IssueDate=datetime.now()
    )
    tenant = Tenant(
        _id='id',
        CreatedAt=datetime.now(),
        UpdatedAt=datetime.now(),
        TenantId='tid',
        IsAcceptBlocksTerms=True,
        IsUseBlocksExclusively=False,
        IsProduction=True,
        Name='name',
        DBName='db',
        ApplicationDomain='app.dom',
        AllowedDomains=['a.com'],
        CookieDomain='cookie.dom',
        IsDisabled=False,
        DbConnectionString='conn',
        TenantSalt='salt',
        JwtTokenParameters=params,
        IsRootTenant=True,
        IsCookieEnable=True,
        IsDomainVerified=True
    )
    assert tenant.tenant_id == 'tid'
    assert tenant.is_accept_blocks_terms is True
    assert tenant.is_use_blocks_exclusively is False
    assert tenant.is_production is True
    assert tenant.name == 'name'
    assert tenant.db_name == 'db'
    assert tenant.application_domain == 'app.dom'
    assert tenant.allowed_domains == ['a.com']
    assert tenant.cookie_domain == 'cookie.dom'
    assert tenant.is_disabled is False
    assert tenant.db_connection_string == 'conn'
    assert tenant.tenant_salt == 'salt'
    assert isinstance(tenant.jwt_token_parameters, JwtTokenParameters)
    assert tenant.is_root_tenant is True
    assert tenant.is_cookie_enable is True
    assert tenant.is_domain_verified is True 


from asyncio import sleep
from api import AiMessage


async def handle_user_created_event(event_data):
    ai_message = AiMessage(**event_data)
    print(f"Handling user created event: {ai_message.message}")
    await sleep(60)



import pytest
from unittest.mock import AsyncMock, patch, MagicMock
from blocks_genesis._middlewares.tenant_middleware import TenantValidationMiddleware
from fastapi import Request

@pytest.mark.asyncio
@patch('blocks_genesis._middlewares.tenant_middleware.get_tenant_service')
@patch('blocks_genesis._middlewares.tenant_middleware.Activity')
@patch('blocks_genesis._middlewares.tenant_middleware.BlocksContextManager')
async def test_dispatch_excluded_paths(mock_ctx_mgr, mock_activity, mock_get_tenant_service):
    middleware = TenantValidationMiddleware(MagicMock())
    request = MagicMock(spec=Request)
    request.url.path = '/docs'
    call_next = AsyncMock(return_value='response')
    result = await middleware.dispatch(request, call_next)
    assert result == 'response'

@pytest.mark.asyncio
@patch('blocks_genesis._middlewares.tenant_middleware.get_tenant_service')
@patch('blocks_genesis._middlewares.tenant_middleware.Activity')
@patch('blocks_genesis._middlewares.tenant_middleware.BlocksContextManager')
async def test_dispatch_missing_tenant(mock_ctx_mgr, mock_activity, mock_get_tenant_service):
    middleware = TenantValidationMiddleware(MagicMock())
    request = MagicMock(spec=Request)
    request.url.path = '/not-excluded'
    request.headers.get.return_value = None
    request.query_params.get.return_value = None
    request.base_url.hostname = 'host'
    tenant_service = mock_get_tenant_service.return_value
    tenant_service.get_tenant_by_domain = AsyncMock(return_value=None)
    call_next = AsyncMock()
    result = await middleware.dispatch(request, call_next)
    assert result.status_code == 404

@pytest.mark.asyncio
@patch('blocks_genesis._middlewares.tenant_middleware.get_tenant_service')
@patch('blocks_genesis._middlewares.tenant_middleware.Activity')
@patch('blocks_genesis._middlewares.tenant_middleware.BlocksContextManager')
async def test_dispatch_valid_tenant(mock_ctx_mgr, mock_activity, mock_get_tenant_service):
    middleware = TenantValidationMiddleware(MagicMock())
    request = MagicMock(spec=Request)
    request.url.path = '/not-excluded'
    request.headers.get.return_value = 'api-key'
    request.query_params.get.return_value = None
    tenant = MagicMock()
    tenant.is_disabled = False
    tenant.is_root_tenant = True
    tenant.tenant_id = 'tid'
    tenant.allowed_domains = []
    tenant.application_domain = 'host'
    tenant_service = mock_get_tenant_service.return_value
    tenant_service.get_tenant = AsyncMock(return_value=tenant)
    call_next = AsyncMock(return_value=MagicMock(status_code=200, headers={}, __class__=MagicMock()))
    result = await middleware.dispatch(request, call_next)
    assert hasattr(result, 'status_code')

@pytest.mark.asyncio
@patch('blocks_genesis._middlewares.tenant_middleware.get_tenant_service')
@patch('blocks_genesis._middlewares.tenant_middleware.Activity')
@patch('blocks_genesis._middlewares.tenant_middleware.BlocksContextManager')
async def test_dispatch_invalid_origin(mock_ctx_mgr, mock_activity, mock_get_tenant_service):
    middleware = TenantValidationMiddleware(MagicMock())
    request = MagicMock(spec=Request)
    request.url.path = '/not-excluded'
    request.headers.get.return_value = 'api-key'
    request.query_params.get.return_value = None
    tenant = MagicMock()
    tenant.is_disabled = False
    tenant.is_root_tenant = True
    tenant.tenant_id = 'tid'
    tenant.allowed_domains = []
    tenant.application_domain = 'host'
    tenant_service = mock_get_tenant_service.return_value
    tenant_service.get_tenant = AsyncMock(return_value=tenant)
    # Patch _is_valid_origin_or_referer to return False
    middleware._is_valid_origin_or_referer = MagicMock(return_value=False)
    call_next = AsyncMock()
    result = await middleware.dispatch(request, call_next)
    assert result.status_code == 406

def test_reject():
    middleware = TenantValidationMiddleware(MagicMock())
    resp = middleware._reject(400, 'msg')
    assert resp.status_code == 400
    assert resp.body

def test_is_valid_origin_or_referer():
    middleware = TenantValidationMiddleware(MagicMock())
    request = MagicMock()
    tenant = MagicMock()
    tenant.allowed_domains = ['a.com']
    tenant.application_domain = 'host'
    request.headers.get.side_effect = lambda k: 'http://a.com' if k == 'origin' else None
    assert middleware._is_valid_origin_or_referer(request, tenant) 


import base64
import hashlib

class CryptoService:
    @staticmethod
    def hash_string(value: str, salt: str = None) -> str:
        """
        Hash a string with optional salt, compatible with C# implementation.
        
        Args:
            value: String to hash
            salt: Optional salt string (None treated as empty string)
            
        Returns:
            Lowercase hex string without separators
        """
        value_bytes = value.encode('utf-8')

        salt_bytes = (salt or "").encode('utf-8')
        salted_value = value_bytes + salt_bytes
        return CryptoService.hash_bytes(salted_value)
    
    @staticmethod  
    def hash_bytes(value: bytes, make_base64: bool = False) -> str:
        """
        Hash bytes with SHA256, compatible with C# implementation.
        
        Args:
            value: Bytes to hash
            make_base64: If True, return base64 encoded string, else hex
            
        Returns:
            Base64 string or lowercase hex string without separators
        """
        hash_bytes = hashlib.sha256(value).digest()
        if make_base64:
            return base64.b64encode(hash_bytes).decode('utf-8')

        return hash_bytes.hex().lower()


from blocks_genesis._utilities.crypto_service import CryptoService
import base64
import hashlib

def test_hash_string_no_salt():
    value = 'test'
    expected = hashlib.sha256(b'test').hexdigest().lower()
    assert CryptoService.hash_string(value) == expected

def test_hash_string_with_salt():
    value = 'test'
    salt = 'salt'
    expected = hashlib.sha256(b'testsalt').hexdigest().lower()
    assert CryptoService.hash_string(value, salt) == expected

def test_hash_bytes_hex():
    value = b'abc'
    expected = hashlib.sha256(value).hexdigest().lower()
    assert CryptoService.hash_bytes(value) == expected

def test_hash_bytes_base64():
    value = b'abc'
    expected = base64.b64encode(hashlib.sha256(value).digest()).decode('utf-8')
    assert CryptoService.hash_bytes(value, make_base64=True) == expected 


import os
import pytest
from blocks_genesis._core.env_vault_config import EnvVaultConfig

def test_get_config_all(monkeypatch):
    monkeypatch.setenv('FOO', 'BAR')
    config = EnvVaultConfig.get_config()
    assert 'FOO' in config
    assert config['FOO'] == 'BAR'

def test_get_config_keys(monkeypatch):
    monkeypatch.setenv('A', '1')
    monkeypatch.setenv('B', '2')
    config = EnvVaultConfig.get_config(['A', 'B'])
    assert config['A'] == '1'
    assert config['B'] == '2'

def test_get_config_missing(monkeypatch):
    monkeypatch.delenv('X', raising=False)
    with pytest.raises(EnvironmentError):
        EnvVaultConfig.get_config(['X']) 


import pytest
from unittest.mock import AsyncMock, patch, MagicMock
from blocks_genesis._core import secret_loader

@pytest.mark.asyncio
@patch('blocks_genesis._core.secret_loader.AzureKeyVault')
@patch('blocks_genesis._core.secret_loader.BlocksSecret')
async def test_load_secrets_and_get_blocks_secret(mock_blocks_secret, mock_vault, monkeypatch):
    # Reset module-level _loaded_secret
    secret_loader._loaded_secret = None
    mock_vault_instance = mock_vault.return_value
    mock_vault_instance.get_secrets = AsyncMock(return_value={'CacheConnectionString': 'foo', 'ServiceName': 'svc'})
    mock_blocks_secret.return_value = MagicMock()
    loader = secret_loader.SecretLoader('svc')
    await loader.load_secrets()
    # Should set _loaded_secret
    assert secret_loader._loaded_secret is not None
    # get_blocks_secret returns the loaded secret
    assert secret_loader.get_blocks_secret() is not None

@pytest.mark.asyncio
@patch('blocks_genesis._core.secret_loader.AzureKeyVault')
@patch('blocks_genesis._core.secret_loader.BlocksSecret')
async def test_load_secrets_already_loaded(mock_blocks_secret, mock_vault):
    secret_loader._loaded_secret = MagicMock()
    loader = secret_loader.SecretLoader('svc')
    await loader.load_secrets()
    # Should not reload
    mock_vault.return_value.get_secrets.assert_not_called()

@pytest.mark.asyncio
@patch('blocks_genesis._core.secret_loader.AzureKeyVault')
@patch('blocks_genesis._core.secret_loader.BlocksSecret')
async def test_load_secrets_error(mock_blocks_secret, mock_vault):
    secret_loader._loaded_secret = None
    mock_vault_instance = mock_vault.return_value
    mock_vault_instance.get_secrets = AsyncMock(side_effect=Exception('fail'))
    loader = secret_loader.SecretLoader('svc')
    with pytest.raises(Exception):
        await loader.load_secrets()

@pytest.mark.asyncio
@patch('blocks_genesis._core.secret_loader.AzureKeyVault')
async def test_close(mock_vault):
    loader = secret_loader.SecretLoader('svc')
    mock_vault_instance = mock_vault.return_value
    mock_vault_instance.close = AsyncMock()
    await loader.close()
    mock_vault_instance.close.assert_awaited()

def test_get_blocks_secret_not_loaded():
    secret_loader._loaded_secret = None
    with pytest.raises(Exception):
        secret_loader.get_blocks_secret() 


import pytest
from unittest.mock import AsyncMock, patch, MagicMock
from blocks_genesis._database.mongo_context import MongoDbContextProvider

@pytest.mark.asyncio
@patch('blocks_genesis._database.mongo_context.get_tenant_service')
@patch('blocks_genesis._database.mongo_context.register')
@patch('blocks_genesis._database.mongo_context.MongoEventSubscriber')
async def test_get_database_and_collection(mock_sub, mock_register, mock_get_tenant_service):
    provider = MongoDbContextProvider()
    mock_tenant_service = mock_get_tenant_service.return_value
    mock_tenant_service.get_db_connection = AsyncMock(return_value=('db', 'conn'))
    with patch('blocks_genesis._database.mongo_context.MongoClient') as mock_client:
        mock_db = MagicMock()
        mock_client.return_value.__getitem__.return_value = mock_db
        db = await provider.get_database('tid')
        assert db is not None
        col = await provider.get_collection('col', 'tid')
        assert col is not None

@patch('blocks_genesis._database.mongo_context.get_tenant_service')
@patch('blocks_genesis._database.mongo_context.register')
@patch('blocks_genesis._database.mongo_context.MongoEventSubscriber')
def test_get_database_by_connection(mock_sub, mock_register, mock_get_tenant_service):
    provider = MongoDbContextProvider()
    with patch('blocks_genesis._database.mongo_context.MongoClient') as mock_client:
        mock_db = MagicMock()
        mock_client.return_value.__getitem__.return_value = mock_db
        db = provider.get_database_by_connection('conn', 'db')
        assert db is not None

@pytest.mark.asyncio
@patch('blocks_genesis._database.mongo_context.get_tenant_service')
@patch('blocks_genesis._database.mongo_context.register')
@patch('blocks_genesis._database.mongo_context.MongoEventSubscriber')
async def test_get_database_missing_tenant_id(mock_sub, mock_register, mock_get_tenant_service):
    provider = MongoDbContextProvider()
    mock_get_tenant_service.return_value.get_db_connection = AsyncMock(return_value=(None, None))
    db = await provider.get_database(None)
    assert db is None

@patch('blocks_genesis._database.mongo_context.get_tenant_service')
@patch('blocks_genesis._database.mongo_context.register')
@patch('blocks_genesis._database.mongo_context.MongoEventSubscriber')
def test_get_database_by_connection_errors(mock_sub, mock_register, mock_get_tenant_service):
    provider = MongoDbContextProvider()
    with pytest.raises(ValueError):
        provider.get_database_by_connection('', 'db')
    with pytest.raises(ValueError):
        provider.get_database_by_connection('conn', '') 


from fastapi import Request, Response
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware
from datetime import datetime
from opentelemetry.trace import StatusCode
from blocks_genesis._auth.blocks_context import BlocksContextManager
from blocks_genesis._lmt.activity import Activity
from blocks_genesis._tenant.tenant import Tenant
from blocks_genesis._tenant.tenant_service import get_tenant_service


class TenantValidationMiddleware(BaseHTTPMiddleware):

    async def dispatch(self, request: Request, call_next):
        excluded_paths = ["/ping", "/swagger/index.html", "/openapi.json"]
        root_path = request.scope.get("root_path", "")

        all_excluded = excluded_paths + [root_path + path for path in excluded_paths]

        if request.url.path in all_excluded:
            return await call_next(request)
        
        try:
            Activity.set_current_properties({
                "http.query": str(dict(request.query_params)),
                "http.headers": str(dict(request.headers))
            })
            
            
            api_key = request.headers.get("x-blocks-key") or request.query_params.get("x-blocks-key")
            tenant: Tenant = None
            tenant_service = get_tenant_service()  # Assuming this function retrieves the tenant service instance

            if not api_key:
                tenant = await tenant_service.get_tenant_by_domain(request.base_url.hostname)
                if not tenant:
                    return self._reject(404, "Not_Found: Application_Not_Found")
            else:
                tenant = await tenant_service.get_tenant(api_key)

            if not tenant or tenant.is_disabled:
                return self._reject(404, "Not_Found: Application_Not_Found")

            if not self._is_valid_origin_or_referer(request, tenant):
                return self._reject(406, "NotAcceptable: Invalid_Origin_Or_Referer")
            

            Activity.set_current_property("baggage.TenantId", tenant.tenant_id)
            Activity.set_current_property("baggage.IsFromCloud", "true" if tenant.is_root_tenant else "false")
            print(f"TenantId set in baggage: {tenant.tenant_id}")
            # Construct and set BlocksContext
            ctx = BlocksContextManager.create(
                tenant_id=tenant.tenant_id,
                roles=[],
                user_id="",
                is_authenticated=False,
                request_uri=request.url.path,
                organization_id="",
                expire_on=datetime.now(),
                email="",
                permissions=[],
                user_name="",
                phone_number="",
                display_name="",
                oauth_token="",
                actual_tenant_id=tenant.tenant_id
            )
            BlocksContextManager.set_context(ctx)
            Activity.set_current_property("SecurityContext", str(ctx.__dict__))
            
            request_size = int(request.headers.get("content-length", 0))

            response = await call_next(request)
            
            body = b"".join([chunk async for chunk in response.body_iterator])
            response_size = len(body)

            Activity.set_current_property("request.size.bytes", request_size)
            Activity.set_current_property("response.size.bytes", response_size)
            Activity.set_current_property("throughput.total.bytes", request_size + response_size)
            
            response = Response(
                content=body,
                status_code=response.status_code,
                headers=dict(response.headers),
                media_type=response.media_type
            )
            
            if not (200 <= response.status_code < 300):
                Activity.set_current_property(StatusCode.ERROR, f"HTTP {response.status_code}")
            Activity.set_current_properties({
                "response.status.code": response.status_code,
                "response.headers": str(dict(response.headers)),
            })
        
        except Exception as e:
            Activity.set_current_status(StatusCode.ERROR, str(e))
            raise
        finally:
            BlocksContextManager.clear_context()

        return response           
    
     
            
    def _reject(self, status: int, message: str) -> Response:
        return JSONResponse(
            status_code=status,
            content={
                "is_success": False,
                "errors": {"message": message}
            }
        )

    def _is_valid_origin_or_referer(self, request: Request, tenant: Tenant) -> bool:
        def extract_domain(url: str) -> str:
            try:
                return url.split("//")[-1].split("/")[0].split(":")[0]
            except:
                return ""

        allowed = [extract_domain(d) for d in tenant.allowed_domains]
        current = extract_domain(request.headers.get("origin") or "") or extract_domain(request.headers.get("referer") or "")
        normalized_app_domain = extract_domain(tenant.application_domain)

        return not current or current == "localhost" or current == normalized_app_domain or current in allowed



from typing import Callable, Any, Type, Dict, Union

class EventRegistry:
    """
    Registry for event handlers, allowing registration and retrieval by event type.
    """
    _handlers: Dict[str, Union[Callable[..., Any], Type[Any]]] = {}

    @classmethod
    def register(cls, event_type: str):
        """
        Decorator to register a handler for a given event type.
        """
        if not isinstance(event_type, str) or not event_type:
            raise ValueError("event_type must be a non-empty string.")

        def decorator(handler: Union[Callable[..., Any], Type[Any]]):
            if event_type in cls._handlers:
                raise KeyError(f"Handler already registered for event type: {event_type}")
            cls._handlers[event_type] = handler
            return handler
        return decorator

    @classmethod
    def resolve(cls, event_type: str) -> Union[Callable[..., Any], Type[Any]]:
        """
        Retrieve the handler registered for the given event type.
        """
        if event_type not in cls._handlers:
            raise ValueError(f"No handler registered for event type: {event_type}")
        return cls._handlers[event_type]




import pytest
from unittest.mock import patch, MagicMock
from blocks_genesis._database.mongo_event_subscriber import MongoEventSubscriber

@patch('blocks_genesis._database.mongo_event_subscriber.Activity')
@patch('blocks_genesis._database.mongo_event_subscriber.BlocksContextManager')
def test_started_and_succeeded_and_failed(mock_ctx_mgr, mock_activity):
    sub = MongoEventSubscriber()
    # Mock event
    event = MagicMock()
    event.command_name = 'find'
    event.request_id = 1
    # started
    activity_instance = MagicMock()
    mock_activity.start.return_value = activity_instance
    sub.started(event)
    assert sub._activities[event.request_id] == activity_instance
    # succeeded
    event.duration_micros = 1000
    sub.succeeded(event)
    activity_instance.set_property.assert_called_with('db.duration_ms', 1.0)
    activity_instance.set_status.assert_called()
    activity_instance.stop.assert_called()
    # failed
    event2 = MagicMock()
    event2.request_id = 2
    event2.command_name = 'insert'
    event2.duration_micros = 2000
    event2.failure = 'failmsg'
    mock_activity2 = MagicMock()
    sub._activities[event2.request_id] = mock_activity2
    sub.failed(event2)
    mock_activity2.set_property.assert_any_call('db.duration_ms', 2.0)
    mock_activity2.set_property.assert_any_call('exception.message', str(event2.failure))
    mock_activity2.set_status.assert_called()
    mock_activity2.stop.assert_called() 


from datetime import datetime
from typing import List, Optional
from pydantic import BaseModel, Field
from enum import IntEnum

from blocks_genesis._entities.base_entity import BaseEntity

class CertificateStorageType(IntEnum):
    AZURE = 1
    FILESYSTEM = 2
    MONGODB = 3

class JwtTokenParameters(BaseModel):
    issuer: Optional[str] = Field(alias="Issuer", default="")
    subject: Optional[str] = Field(alias="Subject", default=None)
    audiences: List[str] = Field(default_factory=list, alias="Audiences")
    public_certificate_path: Optional[str] = Field(alias="PublicCertificatePath", default=None)
    public_certificate_password: Optional[str] = Field(alias="PublicCertificatePassword", default="")
    private_certificate_password: Optional[str] = Field(alias="PrivateCertificatePassword", default="")
    certificate_storage_type: CertificateStorageType = Field(alias="CertificateStorageType", default=CertificateStorageType.AZURE)
    certificate_valid_for_number_of_days: int = Field(alias="CertificateValidForNumberOfDays", default=365)
    issue_date: Optional[datetime] = Field(alias="IssueDate", default=None)

    class Config:
        extra = "ignore"
        validate_by_name = True
        use_enum_values = True

class Tenant(BaseEntity):
    tenant_id: Optional[str] = Field(alias="TenantId", default="")
    is_accept_blocks_terms: Optional[bool] = Field(alias="IsAcceptBlocksTerms", default=False)
    is_use_blocks_exclusively: Optional[bool] = Field(alias="IsUseBlocksExclusively", default=False)
    is_production: Optional[bool] = Field(alias="IsProduction", default=False)
    name: Optional[str] = Field(alias="Name", default="")
    db_name: Optional[str] = Field(alias="DBName", default="")
    application_domain: Optional[str] = Field(alias="ApplicationDomain", default="")
    allowed_domains: List[str] = Field(default_factory=list, alias="AllowedDomains")
    cookie_domain: Optional[str] = Field(alias="CookieDomain", default="")
    is_disabled: Optional[bool] = Field(alias="IsDisabled", default=False)
    db_connection_string: Optional[str] = Field(alias="DbConnectionString", default="")
    tenant_salt: Optional[str] = Field(alias="TenantSalt", default="")
    jwt_token_parameters: Optional[JwtTokenParameters] = Field(alias="JwtTokenParameters", default=None)
    is_root_tenant: Optional[bool] = Field(alias="IsRootTenant", default=False)
    is_cookie_enable: Optional[bool] = Field(alias="IsCookieEnable", default=False)
    is_domain_verified: Optional[bool] = Field(alias="IsDomainVerified", default=False)
    environment: Optional[str] = Field(alias="Environment", default="")
    tenant_group_id: Optional[str] = Field(alias="TenantGroupId", default="")

    class Config:
        extra = "ignore"
        validate_by_name = True



import asyncio
import logging
import sys

from blocks_genesis._core.worker import WorkerConsoleApp
from blocks_genesis._message.message_configuration import AzureServiceBusConfiguration, MessageConfiguration
from test_consumer import handle_user_created_event



logger = logging.getLogger(__name__)
message_config = MessageConfiguration(
    azure_service_bus_configuration=AzureServiceBusConfiguration(
        queues=["ai_queue"],
        topics=[]
    )
)


async def callback():
    print("Callback called")



def main(): 
    app = WorkerConsoleApp("blocks_ai_worker", message_config, {
        "AiMessage": handle_user_created_event
    })
    
    try:
        asyncio.run(app.run(callback))
    except KeyboardInterrupt:
        logger.exception("\n👋 Graceful shutdown by user")
    except Exception as e:
        logger.exception(f"💥 Fatal error: {e}", file=sys.stderr)
        sys.exit(1)

if __name__ == "__main__":
    main()
    
    




import pytest
from unittest.mock import AsyncMock, patch, MagicMock
from blocks_genesis._middlewares.global_exception_middleware import GlobalExceptionHandlerMiddleware
from starlette.requests import Request

@pytest.mark.asyncio
@patch('blocks_genesis._middlewares.global_exception_middleware.Activity')
async def test_dispatch_normal(mock_activity):
    mock_activity.get_trace_id.return_value = 'traceid'
    middleware = GlobalExceptionHandlerMiddleware(MagicMock())
    request = MagicMock(spec=Request)
    call_next = AsyncMock(return_value='response')
    result = await middleware.dispatch(request, call_next)
    assert result == 'response'

@pytest.mark.asyncio
@patch('blocks_genesis._middlewares.global_exception_middleware.Activity')
async def test_dispatch_exception(mock_activity):
    mock_activity.get_trace_id.return_value = 'traceid'
    middleware = GlobalExceptionHandlerMiddleware(MagicMock())
    request = MagicMock(spec=Request)
    call_next = AsyncMock(side_effect=Exception('fail'))
    with patch.object(middleware, 'handle_exception', AsyncMock(return_value='handled')) as mock_handle:
        result = await middleware.dispatch(request, call_next)
        assert result == 'handled'

@pytest.mark.asyncio
@patch('blocks_genesis._middlewares.global_exception_middleware.Activity')
async def test_handle_exception_json_payload(mock_activity):
    mock_activity.get_trace_id.return_value = 'traceid'
    middleware = GlobalExceptionHandlerMiddleware(MagicMock())
    request = MagicMock(spec=Request)
    request.headers.get.return_value = 'application/json'
    request.body = AsyncMock(return_value=b'{"foo": "bar"}')
    result = await middleware.handle_exception(request, Exception('fail'))
    assert result.status_code == 500
    assert b'An error occurred' in result.body

@pytest.mark.asyncio
@patch('blocks_genesis._middlewares.global_exception_middleware.Activity')
async def test_handle_exception_non_json_payload(mock_activity):
    mock_activity.get_trace_id.return_value = 'traceid'
    middleware = GlobalExceptionHandlerMiddleware(MagicMock())
    request = MagicMock(spec=Request)
    request.headers.get.return_value = 'text/plain'
    request.body = AsyncMock(return_value=b'')
    result = await middleware.handle_exception(request, Exception('fail'))
    assert result.status_code == 500

@pytest.mark.asyncio
@patch('blocks_genesis._middlewares.global_exception_middleware.Activity')
async def test_handle_exception_body_too_long(mock_activity):
    mock_activity.get_trace_id.return_value = 'traceid'
    middleware = GlobalExceptionHandlerMiddleware(MagicMock())
    request = MagicMock(spec=Request)
    request.headers.get.return_value = 'application/json'
    request.body = AsyncMock(return_value=b'a' * 2000)
    result = await middleware.handle_exception(request, Exception('fail'))
    assert result.status_code == 500 


from blocks_genesis._core.blocks_secret import BlocksSecret

def test_blocks_secret_defaults():
    secret = BlocksSecret()
    assert secret.CacheConnectionString == ""
    assert secret.ServiceName == ""

def test_blocks_secret_fields():
    secret = BlocksSecret(CacheConnectionString="foo", ServiceName="svc")
    assert secret.CacheConnectionString == "foo"
    assert secret.ServiceName == "svc" 


import asyncio
import collections
import json
import logging
import types
from typing import Dict, List, Optional
from azure.servicebus.aio import ServiceBusClient, ServiceBusReceiver
from azure.servicebus import ServiceBusReceivedMessage
from opentelemetry import trace
from opentelemetry.trace import Status, StatusCode, SpanKind
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
from blocks_genesis._auth.blocks_context import BlocksContextManager
from blocks_genesis._core.secret_loader import get_blocks_secret
from blocks_genesis._message.consumer import Consumer
from blocks_genesis._message.event_message import EventMessage
from blocks_genesis._message.message_configuration import MessageConfiguration

logger = logging.getLogger(__name__)


class AzureMessageWorker:
    def __init__(self, message_config: MessageConfiguration):
        self._logger = logger
        self._message_config = message_config
        self._consumer = Consumer()
        self._service_bus_client: Optional[ServiceBusClient] = None
        self._receivers: List[ServiceBusReceiver] = []
        self._active_message_renewals: Dict[str, asyncio.Event] = {}
        self._tracer = trace.get_tracer(__name__)

    def initialize(self):
        connection = (
            self._message_config.connection
            or get_blocks_secret().MessageConnectionString
        )
        if not connection:
            self._logger.error("Connection string missing")
            raise ValueError("Connection string missing")
        self._service_bus_client = ServiceBusClient.from_connection_string(connection)
        self._logger.info("Service Bus Client initialized")

    async def stop(self):
        for message_id, event in self._active_message_renewals.items():
            event.set()
        self._active_message_renewals.clear()

        for receiver in self._receivers:
            try:
                await receiver.close()
            except Exception as ex:
                self._logger.error("Error closing receiver: %s", ex)
        self._receivers.clear()

        if self._service_bus_client:
            await self._service_bus_client.close()

        self._logger.info("Worker stopped")

    async def run(self):
        if not self._service_bus_client:
            raise ValueError("Service Bus Client is not initialized")

        receiver_tasks = []

        for queue_name in self._message_config.azure_service_bus_configuration.queues:
            receiver = self._service_bus_client.get_queue_receiver(
                queue_name=queue_name,
                prefetch_count=self._message_config.azure_service_bus_configuration.queue_prefetch_count,
            )
            self._receivers.append(receiver)
            receiver_tasks.append(self.safe_receiver_wrapper(receiver, queue_name))

        for topic_name in self._message_config.azure_service_bus_configuration.topics:
            subscription_name = self._message_config.subscription_name.get(
                topic_name, "default-subscription"
            )
            receiver = self._service_bus_client.get_subscription_receiver(
                topic_name=topic_name,
                subscription_name=subscription_name,
                prefetch_count=self._message_config.azure_service_bus_configuration.topic_prefetch_count,
            )
            self._receivers.append(receiver)
            receiver_tasks.append(self.safe_receiver_wrapper(receiver, topic_name))

        self._logger.info("Receivers started")
        await asyncio.gather(*receiver_tasks, return_exceptions=True)

    async def safe_receiver_wrapper(self, receiver: ServiceBusReceiver, name: str):
        try:
            await self.process_receiver(receiver)
        except Exception as ex:
            self._logger.error(
                "Receiver task for %s crashed but will be ignored: %s", name, ex
            )

    async def process_receiver(self, receiver: ServiceBusReceiver):
        async with receiver:
            async for message in receiver:
                try:
                    await self.message_handler(receiver, message)
                except Exception as ex:
                    self._logger.error(
                        "Error in message_handler, skipping message: %s", ex
                    )
                    try:
                        await receiver.abandon_message(message)
                    except Exception as abandon_ex:
                        self._logger.warning(
                            "Failed to abandon message: %s", abandon_ex
                        )

    def decode_app_properties(self, properties):
        if not properties:
            return {}
        return {
            (k.decode("utf-8") if isinstance(k, bytes) else k): (
                v.decode("utf-8") if isinstance(v, bytes) else v
            )
            for k, v in properties.items()
        }

    async def message_handler(
        self, receiver: ServiceBusReceiver, message: ServiceBusReceivedMessage
    ):
        message_id = message.message_id
        self._logger.info("Received message: %s", message_id)

        app_props = self.decode_app_properties(message.application_properties)
        trace_id = app_props.get("TraceId", "")
        span_id = app_props.get("SpanId", "")
        tenant_id = app_props.get("TenantId", "")
        security_context_raw = app_props.get("SecurityContext", "")
        baggage_str = app_props.get("Baggage", "{}")

        if security_context_raw:
            try:
                sc = json.loads(security_context_raw)
                BlocksContextManager.set_context(BlocksContextManager.create(**sc))
            except Exception as ctx_err:
                self._logger.warning("Invalid security context: %s", ctx_err)

        cancellation_event = asyncio.Event()
        self._active_message_renewals[message_id] = cancellation_event
        renewal_task = asyncio.create_task(
            self.start_auto_renewal_task(message, receiver, cancellation_event)
        )
        
        try:
            context = (
                TraceContextTextMapPropagator().extract(
                    {"traceparent": f"00-{trace_id}-{span_id}-01"}
                )
                if trace_id and span_id
                else None
            )

            with self._tracer.start_as_current_span(
                "process.messaging.azure.service.bus",
                context=context,
                kind=SpanKind.CONSUMER,
            ) as span:
                span.set_attribute("messaging.system", "azure.servicebus")
                span.set_attribute("message.id", message_id)
                span.set_attribute("SecurityContext", security_context_raw)
                span.set_attribute("message.body", str(message))
                span.set_attribute("baggage.TenantId", tenant_id)

                try:
                    baggages = json.loads(baggage_str)
                    for key, value in baggages.items():
                        span.set_attribute(f"baggage.{key}", value)
                except json.JSONDecodeError:
                    self._logger.warning("Invalid baggage JSON")

                start_time = asyncio.get_event_loop().time()
                if message.body is None:
                    d = "{}"
                else:
                    body_bytes = (
                        b"".join(message.body)
                        if isinstance(
                            message.body,
                            (types.GeneratorType, collections.abc.Iterable),
                        )
                        else message.body
                    )
                    d = (
                        body_bytes.decode("utf-8")
                        if isinstance(body_bytes, (bytes, bytearray))
                        else str(body_bytes)
                    )

                msg = EventMessage(**json.loads(d))
                await self._consumer.process_message(msg.type, msg.body)
                processing_time = (asyncio.get_event_loop().time() - start_time) * 1000
                self._logger.info(
                    "Processed message %s in %.1fms", message_id, processing_time
                )

                span.set_attribute("response", "Successfully Completed")
                span.set_status(Status(StatusCode.OK, "Message processed successfully"))

                await receiver.complete_message(message)
                self._logger.info("Message %s completed", message_id)

        except Exception as ex:
            self._logger.error(
                "Processing failed for message %s: %s", message_id, ex
            )
            try:
                await receiver.abandon_message(message)
            except Exception as abandon_ex:
                self._logger.warning("Abandon message failed: %s", abandon_ex)
            raise
        finally:
            cancellation_event.set()
            self._active_message_renewals.pop(message_id, None)
            renewal_task.cancel()
            BlocksContextManager.clear_context()

    async def start_auto_renewal_task(
        self,
        message: ServiceBusReceivedMessage,
        receiver: ServiceBusReceiver,
        cancellation_event: asyncio.Event,
    ):
        message_id = message.message_id
        start_time = asyncio.get_event_loop().time()
        renewal_count = 0
        renewal_interval = (
            self._message_config.azure_service_bus_configuration.message_lock_renewal_interval_seconds
        )
        max_processing_time = (
            self._message_config.azure_service_bus_configuration.max_message_processing_time_in_minutes
            * 60
        )

        try:
            while not cancellation_event.is_set():
                try:
                    await asyncio.wait_for(
                        cancellation_event.wait(), timeout=renewal_interval
                    )
                except asyncio.TimeoutError:
                    pass  # normal

                if cancellation_event.is_set():
                    break

                processing_time = asyncio.get_event_loop().time() - start_time
                if processing_time > max_processing_time:
                    self._logger.warning(
                        "Message %s exceeded max time; stopping lock renewal",
                        message_id,
                    )
                    break

                try:
                    await receiver.renew_message_lock(message)
                    renewal_count += 1
                    self._logger.info(
                        "Renewed lock for message %s (%s)", message_id, renewal_count
                    )
                except Exception as ex:
                    self._logger.warning(
                        "Lock renewal failed for %s: %s", message_id, ex
                    )
                    cancellation_event.set()
                    
        except asyncio.CancelledError:
            self._logger.info("Auto-renewal cancelled for %s", message_id)
        except Exception as ex:
            self._logger.error("Auto-renewal error for %s: %s", message_id, ex)
        finally:
            self._logger.info(
                "Auto-renewal finished for %s after %s renewals",
                message_id,
                renewal_count,
            )


import pytest
from unittest.mock import AsyncMock, patch, MagicMock
from blocks_genesis._tenant import tenant_service

@pytest.mark.asyncio
@patch('blocks_genesis._tenant.tenant_service.get_blocks_secret')
@patch('blocks_genesis._tenant.tenant_service.CacheProvider')
@patch('blocks_genesis._tenant.tenant_service.AsyncIOMotorClient')
async def test_tenant_service_init(mock_motor, mock_cache_provider, mock_get_secret):
    mock_get_secret.return_value.DatabaseConnectionString = 'conn'
    mock_get_secret.return_value.RootDatabaseName = 'rootdb'
    mock_cache_provider.get_client.return_value = MagicMock()
    service = tenant_service.TenantService()
    assert service.database is not None

@pytest.mark.asyncio
@patch('blocks_genesis._tenant.tenant_service.TenantService._load_tenants', new_callable=AsyncMock)
@patch('blocks_genesis._tenant.tenant_service.CacheProvider')
@patch('blocks_genesis._tenant.tenant_service.get_blocks_secret')
@patch('blocks_genesis._tenant.tenant_service.AsyncIOMotorClient')
async def test_initialize(mock_motor, mock_get_secret, mock_cache_provider, mock_load_tenants):
    mock_get_secret.return_value.DatabaseConnectionString = 'conn'
    mock_get_secret.return_value.RootDatabaseName = 'rootdb'
    mock_cache_provider.get_client.return_value = MagicMock()
    service = tenant_service.TenantService()
    await service.initialize()
    assert service._initialized is True

@pytest.mark.asyncio
@patch('blocks_genesis._tenant.tenant_service.TenantService._load_tenant_from_db', new_callable=AsyncMock)
@patch('blocks_genesis._tenant.tenant_service.get_blocks_secret')
@patch('blocks_genesis._tenant.tenant_service.CacheProvider')
@patch('blocks_genesis._tenant.tenant_service.AsyncIOMotorClient')
async def test_get_tenant(mock_motor, mock_cache_provider, mock_get_secret, mock_load_tenant):
    mock_get_secret.return_value.DatabaseConnectionString = 'conn'
    mock_get_secret.return_value.RootDatabaseName = 'rootdb'
    mock_cache_provider.get_client.return_value = MagicMock()
    service = tenant_service.TenantService()
    # Not found in cache, triggers db load
    mock_load_tenant.return_value = MagicMock(tenant_id='tid')
    tenant = await service.get_tenant('tid')
    assert tenant is not None
    # Found in cache
    tenant2 = await service.get_tenant('tid')
    assert tenant2 is not None

@pytest.mark.asyncio
@patch('blocks_genesis._tenant.tenant_service.get_blocks_secret')
@patch('blocks_genesis._tenant.tenant_service.CacheProvider')
@patch('blocks_genesis._tenant.tenant_service.AsyncIOMotorClient')
async def test_get_tenant_by_domain(mock_motor, mock_cache_provider, mock_get_secret):
    mock_get_secret.return_value.DatabaseConnectionString = 'conn'
    mock_get_secret.return_value.RootDatabaseName = 'rootdb'
    mock_cache_provider.get_client.return_value = MagicMock()
    service = tenant_service.TenantService()
    mock_db = MagicMock()
    mock_db.__getitem__.return_value.find_one = AsyncMock(return_value={'TenantId': 'tid', 'JwtTokenParameters': {}})
    service.database = mock_db
    tenant = await service.get_tenant_by_domain('domain')
    assert tenant is not None

@pytest.mark.asyncio
@patch('blocks_genesis._tenant.tenant_service.TenantService.get_tenant', new_callable=AsyncMock)
async def test_get_db_connection(mock_get_tenant):
    service = tenant_service.TenantService.__new__(tenant_service.TenantService)
    mock_get_tenant.return_value = MagicMock(db_name='db', db_connection_string='conn')
    service.get_tenant = mock_get_tenant
    db, conn = await service.get_db_connection('tid')
    assert db == 'db'
    assert conn == 'conn'

@pytest.mark.asyncio
def test__load_tenants():
    service = tenant_service.TenantService.__new__(tenant_service.TenantService)
    mock_db = MagicMock()
    mock_cursor = AsyncMock()
    mock_cursor.__aiter__.return_value = [{"TenantId": "tid", "JwtTokenParameters": {}}]
    mock_db.__getitem__.return_value.find.return_value = mock_cursor
    service.database = mock_db
    service._tenant_cache = {}
    import asyncio
    async def run():
        await service._load_tenants()
        assert 'tid' in service._tenant_cache
    asyncio.run(run())

@pytest.mark.asyncio
def test__load_tenant_from_db():
    service = tenant_service.TenantService.__new__(tenant_service.TenantService)
    mock_db = MagicMock()
    mock_db.__getitem__.return_value.find_one = AsyncMock(return_value={'TenantId': 'tid', 'JwtTokenParameters': {}})
    service.database = mock_db
    import asyncio
    async def run():
        tenant = await service._load_tenant_from_db('tid')
        assert tenant is not None
    asyncio.run(run())

@pytest.mark.asyncio
def test__subscribe_to_updates():
    service = tenant_service.TenantService.__new__(tenant_service.TenantService)
    mock_cache = MagicMock()
    mock_cache.subscribe_async = AsyncMock()
    service.cache = mock_cache
    service._update_channel = 'chan'
    service._handle_update = AsyncMock()
    import asyncio
    async def run():
        await service._subscribe_to_updates()
        mock_cache.subscribe_async.assert_awaited()
    asyncio.run(run())

@pytest.mark.asyncio
def test__handle_update():
    service = tenant_service.TenantService.__new__(tenant_service.TenantService)
    service._load_tenants = AsyncMock()
    import asyncio
    async def run():
        await service._handle_update('chan', 'msg')
        service._load_tenants.assert_awaited()
    asyncio.run(run())

def test_get_tenant_service_and_initialize(monkeypatch):
    # Reset global
    tenant_service._tenant_service = None
    with pytest.raises(RuntimeError):
        tenant_service.get_tenant_service()
    class DummyService:
        async def initialize(self):
            pass
    monkeypatch.setattr(tenant_service, 'TenantService', DummyService)
    import asyncio
    asyncio.run(tenant_service.initialize_tenant_service()) 